# Author: Kshiteesh Hegde, kshiteesh.hegde@gmail.com
# If there is a different author for particular methods, they are mentioned there


import networkx as nx
import numpy as np
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import Dropout
from keras.layers import Flatten
from keras.layers.convolutional import Conv2D
from keras.layers.convolutional import MaxPooling2D
from keras.utils import np_utils
from keras.models import model_from_json
from keras import backend as k
import random
import config
import utils
k.set_image_dim_ordering('th')
np.random.seed(7)


def build_model(num_nodes, num_classes):
    model = Sequential()
    if num_nodes < 8:
        model.add(Conv2D(32, (3, 3), padding='valid', input_shape=(1, num_nodes, num_nodes), activation='relu'))
    else:
        model.add(Conv2D(32, (5, 5), padding='valid', input_shape=(1, num_nodes, num_nodes), activation='relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.2))
    model.add(Flatten())
    model.add(Dense(128, activation='relu'))
    model.add(Dense(num_classes, activation='softmax'))
    model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
    # Save the model
    with open('models/model_' + config.dataset + '_' + str(num_nodes) + '.json', 'w') as f:
        f.write(model.to_json())
    model.save_weights('models/model_' + config.dataset + '_' + str(num_nodes) + '.h5')
    return model


def classify(x, y, nodes):
    n_samples = x.shape[0]
    perm = np.random.permutation(x.shape[0])
    x = np.take(x, perm, axis=0)
    y = np.take(y, perm)
    y = np_utils.to_categorical(y).astype('int32')

    x_train = x[:n_samples // 3, :, :]
    x_valid = x[n_samples // 3:(2 * n_samples) // 3, :, :]
    x_test = x[(2 * n_samples) // 3:, :, :]

    y_train = y[:n_samples // 3, :]
    y_valid = y[n_samples // 3:(2 * n_samples) // 3, :]
    y_test = y[(2 * n_samples) // 3:, :]
    actuals = y_test.argmax(1).astype('int32')

    # Reshape for Keras
    x_train = x_train.reshape(x_train.shape[0], 1, nodes, nodes).astype('float32')
    x_valid = x_valid.reshape(x_valid.shape[0], 1, nodes, nodes).astype('float32')
    x_test = x_test.reshape(x_test.shape[0], 1, nodes, nodes).astype('float32')

    # Build & Train/Load the model
    # Delete the model files if you want to train afresh
    num_classes = y_test.shape[1]
    try:
        with open('models/model_' + config.dataset + '_' + str(nodes) + '.json', 'r') as f:
            model = model_from_json(f.read())
            model.load_weights('models/model_' + config.dataset + '_' + str(nodes) + '.h5')
    except IOError:
        model = build_model(nodes, num_classes)
        model.fit(x_train, y_train, validation_data=(x_valid, y_valid), epochs=1, batch_size=200, verbose=2)

    # Final evaluation of the model
    preds = model.predict(x_test)
    preds = preds.argmax(1).astype('int32')
    return sum(1 for i in range(len(actuals)) if actuals[i] == preds[i]) * 1.0 / len(actuals)


def edge_swap(graph, nswap):
    # Tweaked version of the networkx version. see: networkx/algorithms/swap.py for authorship details
    tries = 0
    swapcount = 0
    keys, degrees = zip(*graph.degree().items())  # keys, degree
    cdf = nx.utils.cumulative_distribution(degrees)  # cdf of degree
    while swapcount < nswap:
        # if random.random() < 0.5: continue # trick to avoid periodicities?
        # pick two random edges without creating edge list
        # choose source node indices from discrete distribution
        (ui, xi) = nx.utils.discrete_sequence(2, cdistribution=cdf)
        if ui == xi:
            continue  # same source, skip
        u = keys[ui]  # convert index to label
        x = keys[xi]
        # choose target uniformly from neighbors
        v = random.choice(list(graph[u]))
        y = random.choice(list(graph[x]))
        if v == y:
            continue  # same target, skip
        if (x not in graph[u]) and (y not in graph[v]):
            graph.add_edge(u, x)
            graph.add_edge(v, y)
            graph.remove_edge(u, v)
            graph.remove_edge(x, y)
            swapcount += 1
        tries += 1
    return graph


def main():
    results_file = 'results/' + config.dataset + '_' + str(config.min_subgraph_size) + '_' + \
                   str(config.max_subgraph_size) + '_swaps_' + str(config.delta) + '.txt'
    num_samples = config.num_sg
    graph = nx.read_edgelist('data/' + config.dataset + config.extension)
    if config.delta != 'inf':
        swaps = len(graph.edges()) * config.delta * 1.0 / 100
        graph_random = nx.read_edgelist('data/' + config.dataset + config.extension)
        graph_random = edge_swap(graph_random, swaps)
    else:
        # graph_random = nx.random_degree_sequence_graph(graph.degree().values())
        graph_random = nx.havel_hakimi_graph(graph.degree().values())
    accuracy_scores = [0 for _ in range(config.min_subgraph_size, config.max_subgraph_size)]
    wf = open(results_file, 'w')
    for r in range(config.runs):
        print('Run: ' + str(r + 1) + ' of ' + config.dataset + ' for delta = ' + str(config.delta))
        for j, n in enumerate(range(config.min_subgraph_size, config.max_subgraph_size)):
            print('kappa = ' + str(n) + ' for ' + config.dataset)
            x = np.empty([2 * num_samples, n, n])
            y = np.empty([2 * num_samples])
            sample = 0
            while sample < num_samples:
                # Perform random walks and extract subgraphs from original and perturbed versions
                g = utils.sample_random_walk(graph, n)
                gr = utils.sample_random_walk(graph_random, n)
                # If kappa nodes are not present, try again
                if nx.number_of_nodes(g) != n or nx.number_of_nodes(gr) != n:
                    continue
                # Structure the subgraph
                adj_g = utils.get_adj(g)
                adj_gr = utils.get_adj(gr)
                x[sample], y[sample] = adj_g, 0
                x[num_samples + sample], y[num_samples + sample] = adj_gr, 1
                sample += 1
            # Classification:
            score = classify(x, y, n)
            accuracy_scores[j] += score
    for i in range(len(accuracy_scores)):
        wf.write(str(accuracy_scores[i] / config.runs))
        wf.write('\n')
    wf.close()


if __name__ == '__main__':
    main()
